@@ -66,7 +66,7 @@
  * at once, potentially across multiple navigation views.
  *
  * Set [property@HeaderBar:show-back-button] to `FALSE` to disable this behavior
- * if it's unwanted.
+ * in rare scenarios where it's unwanted.
  *
  * `AdwHeaderBar` will also display the title of the `AdwNavigationPage` it's
  * placed into, so most applications shouldn't need to customize it at all.
@@ -395,6 +395,20 @@ set_remove_on_pop (AdwNavigationPage *self,
   priv->remove_on_pop = remove_on_pop;
 }
 
+static void
+adw_navigation_page_realize (GtkWidget *widget)
+{
+  AdwNavigationPage *self = ADW_NAVIGATION_PAGE (widget);
+  AdwNavigationPagePrivate *priv = adw_navigation_page_get_instance_private (self);
+
+  GTK_WIDGET_CLASS (adw_navigation_page_parent_class)->realize (widget);
+
+  if (!(priv->title && *priv->title) && !priv->child_view) {
+    g_warning ("AdwNavigationPage %p is missing a title. To hide a header bar " \
+               "title, consider using AdwHeaderBar:show-title instead.", self);
+  }
+}
+
 static void
 adw_navigation_page_dispose (GObject *object)
 {
@@ -492,6 +506,7 @@ adw_navigation_page_class_init (AdwNavigationPageClass *klass)
   object_class->get_property = adw_navigation_page_get_property;
   object_class->set_property = adw_navigation_page_set_property;
 
+  widget_class->realize = adw_navigation_page_realize;
   widget_class->compute_expand = adw_widget_compute_expand;
 
   /**
@@ -1556,7 +1571,7 @@ adw_navigation_view_root (GtkWidget *widget)
 
   GTK_WIDGET_CLASS (adw_navigation_view_parent_class)->root (widget);
 
-  parent_page = adw_widget_get_ancestor_same_native (widget, ADW_TYPE_NAVIGATION_PAGE);
+  parent_page = adw_widget_get_ancestor (widget, ADW_TYPE_NAVIGATION_PAGE, TRUE, TRUE);
 
   if (parent_page)
     set_child_view (ADW_NAVIGATION_PAGE (parent_page), self);
@@ -1567,7 +1582,7 @@ adw_navigation_view_unroot (GtkWidget *widget)
 {
   GtkWidget *parent_page;
 
-  parent_page = adw_widget_get_ancestor_same_native (widget, ADW_TYPE_NAVIGATION_PAGE);
+  parent_page = adw_widget_get_ancestor (widget, ADW_TYPE_NAVIGATION_PAGE, TRUE, TRUE);
 
   if (parent_page)
     set_child_view (ADW_NAVIGATION_PAGE (parent_page), NULL);
